"""Module to deal with orbit acquisition."""
import logging as _log

import numpy as _np

from ..diagbeam.bpm.csdev import Const as _CSBPM
from ..envars import VACA_PREFIX as LL_PREF
from ..epics import PV as _PV
from ..search import HLTimeSearch as _HLTimesearch
from ..timesys.csdev import Const as _TIConst
from .base_class import BaseTimingConfig as _BaseTimingConfig

TIMEOUT = 0.05


class BPM(_BaseTimingConfig):
    """."""

    def __init__(self, name, callback=None):
        """."""
        super().__init__(name[:2], callback)
        self.has_news = True

        self._name = name
        self._orb_conv_unit = self._csorb.ORBIT_CONVERSION_UNIT
        pvpref = LL_PREF + ("-" if LL_PREF else "") + self._name + ":"
        opt = {"connection_timeout": TIMEOUT, "auto_monitor": False}
        self._pvs['poskx'] = _PV(pvpref + "PosKx-RB", **opt)
        self._pvs['posky'] = _PV(pvpref + "PosKy-RB", **opt)
        self._pvs['ksum'] = _PV(pvpref + "PosKsum-RB", **opt)
        self._pvs['polyx'] = _PV(pvpref + "GEN_PolyXArrayCoeff-RB", **opt)
        self._pvs['polyy'] = _PV(pvpref + "GEN_PolyYArrayCoeff-RB", **opt)
        self._pvs['arraya'] = _PV(pvpref + "GENAmplAData", **opt)
        self._pvs['arrayb'] = _PV(pvpref + "GENAmplBData", **opt)
        self._pvs['arrayc'] = _PV(pvpref + "GENAmplCData", **opt)
        self._pvs['arrayd'] = _PV(pvpref + "GENAmplDData", **opt)
        self._pvs['arrayx'] = _PV(pvpref + "GENPosXData", **opt)
        self._pvs['arrayy'] = _PV(pvpref + "GENPosYData", **opt)
        self._pvs['arrays'] = _PV(pvpref + "GENSumData", **opt)
        opt.pop("auto_monitor")
        self._pvs['offsetx'] = _PV(pvpref + "PosXOffset-RB", **opt)
        self._pvs['offsety'] = _PV(pvpref + "PosYOffset-RB", **opt)
        self._config_ok_vals = {
            "SwMode": _CSBPM.SwModes.switching,
            "ACQChannel": _CSBPM.AcqChan.ADCSwp,
            "ACQShots": 1,
            "ACQUpdateTime": 0.001,
            "ACQSamplesPre": 0,
            "ACQSamplesPost": 382,
            "ACQTriggerEvent": _CSBPM.AcqEvents.Stop,
            "ACQTrigger": _CSBPM.AcqTrigTyp.External,
            "ACQTriggerRep": _CSBPM.AcqRepeat.Repetitive,
            "ACQDataTrigChan": _CSBPM.AcqChan.ADC,
            "TbTPhaseSyncEn": _CSBPM.DsblEnbl.disabled,  # Enable TbT sync
            "FOFBPhaseSyncEn": _CSBPM.DsblEnbl.disabled,  # Enable FOFB sync
            "FAcqPhaseSyncEn": _CSBPM.DsblEnbl.disabled,  # Enable FAcq sync
            "MonitPhaseSyncEn": _CSBPM.DsblEnbl.disabled,  # Enable Monit sync
            "TbTDataMaskEn": _CSBPM.DsblEnbl.disabled,  # Enable use of mask
            "TbTDataMaskSamplesBeg": 0,
            "TbTDataMaskSamplesEnd": 0,
            "XYPosCal": _CSBPM.DsblEnbl.enabled,
            "SumPosCal": _CSBPM.DsblEnbl.enabled,
            "SwPhaseSyncEn": _CSBPM.DsblEnbl.enabled,  # Enable Switching sync
            "TestDataEn": _CSBPM.DsblEnbl.disabled,
        }
        if self._name.sec in {'SI', 'BO'}:
            self._config_ok_vals["ACQChannel"] = _CSBPM.AcqChan.TbT
        pvs = {
            "SwMode": "SwMode-Sel",
            "ACQChannel": "GENChannel-Sel",
            "ACQShots": "GENShots-SP",
            "ACQUpdateTime": "GENUpdateTime-SP",
            "ACQSamplesPre": "GENSamplesPre-SP",
            "ACQSamplesPost": "GENSamplesPost-SP",
            "ACQTriggerEvent": "GENTriggerEvent-Cmd",
            "ACQTrigger": "GENTrigger-Sel",
            "ACQTriggerRep": "GENTriggerRep-Sel",
            "ACQDataTrigChan": "GENDataTrigChan-Sel",
            "TbTPhaseSyncEn": "TbTPhaseSyncEn-Sel",  # Enable TbT sync
            "FOFBPhaseSyncEn": "FOFBPhaseSyncEn-Sel",  # Enable FOFB sync
            "FAcqPhaseSyncEn": "FAcqPhaseSyncEn-Sel",  # Enable FAcq sync
            "MonitPhaseSyncEn": "MonitPhaseSyncEn-Sel",  # Enable Monit sync
            "TbTDataMaskEn": "TbTDataMaskEn-Sel",  # Enable use of mask
            "TbTDataMaskSamplesBeg": "TbTDataMaskSamplesBeg-SP",
            "TbTDataMaskSamplesEnd": "TbTDataMaskSamplesEnd-SP",
            "XYPosCal": "XYPosCal-Sel",
            "SumPosCal": "SumPosCal-Sel",
            "SwPhaseSyncEn": "SwPhaseSyncEn-Sel",  # Enable Switching sync
            "TestDataEn": "TestDataEn-Sel",
        }
        self._config_pvs_sp = {
            k: _PV(pvpref + v, **opt) for k, v in pvs.items()
        }
        pvs = {
            "INFOClkFreq": "INFOClkFreq-RB",
            "INFOHarmonicNumber": "INFOHarmonicNumber-RB",
            "INFOTbTRate": "INFOTbTRate-RB",
            "INFOFOFBRate": "INFOFOFBRate-RB",
            "INFOMONITRate": "INFOMONITRate-RB",
            "INFOFAcqRate": "INFOFAcqRate-RB",
            "SwMode": "SwMode-Sts",
            "ACQChannel": "GENChannel-Sts",
            "ACQShots": "GENShots-RB",
            "ACQUpdateTime": "GENUpdateTime-RB",
            "ACQSamplesPre": "GENSamplesPre-RB",
            "ACQSamplesPost": "GENSamplesPost-RB",
            "ACQTriggerEvent": "GENTriggerEvent-Cmd",
            "ACQCount": "GENCount-Mon",
            "ACQStatus": "GENStatus-Mon",
            "ACQTrigger": "GENTrigger-Sts",
            "ACQTriggerRep": "GENTriggerRep-Sts",
            "ACQDataTrigChan": "GENDataTrigChan-Sts",
            "TbTPhaseSyncEn": "TbTPhaseSyncEn-Sts",
            "FOFBPhaseSyncEn": "FOFBPhaseSyncEn-Sts",
            "FAcqPhaseSyncEn": "FAcqPhaseSyncEn-Sts",
            "MonitPhaseSyncEn": "MonitPhaseSyncEn-Sts",
            "TbTDataMaskEn": "TbTDataMaskEn-Sts",
            "TbTDataMaskSamplesBeg": "TbTDataMaskSamplesBeg-RB",
            "TbTDataMaskSamplesEnd": "TbTDataMaskSamplesEnd-RB",
            "XYPosCal": "XYPosCal-Sts",
            "SumPosCal": "SumPosCal-Sts",
            "SwPhaseSyncEn": "SwPhaseSyncEn-Sel",
            "TestDataEn": "TestDataEn-Sel",
        }
        self._config_pvs_rb = {
            k: _PV(pvpref + v, **opt) for k, v in pvs.items()
        }
        self._config_pvs_rb["ACQCount"].auto_monitor = True
        self._config_pvs_rb["ACQCount"].add_callback(
            self._reset_has_news
        )

    @property
    def name(self):
        """."""
        return self._name

    @property
    def is_ok(self):
        """."""
        if not super().is_ok:
            return False

        pvobj = self._config_pvs_rb["ACQStatus"]
        stts = _CSBPM.AcqStates
        okay = pvobj.value not in {
            stts.Error,
            stts.No_Memory,
            stts.No_Samples,
            stts.Too_Many_Samples,
            stts.Overflow,
            stts.Bad_Post_Samples,
        }

        if self._config_ok_vals["ACQTriggerEvent"] == _CSBPM.AcqEvents.Start:
            okay &= pvobj.value != stts.Idle
        else:
            okay &= pvobj.value not in {
                stts.Waiting,
                stts.External_Trig,
                stts.Data_Trig,
                stts.Software_Trig,
                stts.Acquiring,
            }
        if not okay:
            msg = "ERR: Error in {0:s}".format(pvobj.pvname)
            self.run_callbacks("Log-Mon", msg)
            _log.warning(msg[5:])
        return okay

    @property
    def switching_mode(self):
        """."""
        pvobj = self._config_pvs_rb["SwMode"]
        if pvobj.connected:
            return pvobj.value
        return None

    @switching_mode.setter
    def switching_mode(self, val):
        """."""
        pvobj = self._config_pvs_sp["SwMode"]
        self._config_ok_vals["SwMode"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def adcfreq(self):
        """."""
        defv = 218446014.0 if self._csorb.acc == "BO" else 220870069.0
        pvobj = self._config_pvs_rb["INFOClkFreq"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def tbtrate(self):
        """."""
        defv = 362 if self._csorb.acc == "BO" else 382
        pvobj = self._config_pvs_rb["INFOTbTRate"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def tbtperiod(self):
        """."""
        return self.tbtrate / self.adcfreq

    @property
    def fofbrate(self):
        """."""
        defv = (362 if self._csorb.acc == "BO" else 382) * 24
        pvobj = self._config_pvs_rb["INFOFOFBRate"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def fofbperiod(self):
        """."""
        return self.fofbrate / self.adcfreq

    @property
    def monitrate(self):
        """."""
        defv = (362 if self._csorb.acc == "BO" else 382) * 59904
        pvobj = self._config_pvs_rb["INFOMONITRate"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def monitperiod(self):
        """."""
        return self.monitrate / self.adcfreq

    @property
    def facqrate(self):
        """."""
        defv = (362 if self._csorb.acc == "BO" else 382) * 603
        pvobj = self._config_pvs_rb["INFOFAcqRate"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def facqperiod(self):
        """."""
        return self.facqrate / self.adcfreq

    @property
    def poskx(self):
        """."""
        defv = 1
        pvobj = self._pvs['poskx']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def posky(self):
        """."""
        defv = 1
        pvobj = self._pvs['posky']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def polyx(self):
        """."""
        pvobj = self._pvs['polyx']
        if pvobj.connected:
            val = pvobj.value
            if val is not None:
                return val
        defv = _np.zeros(15, dtype=float)
        defv[0] = 1
        return defv

    @property
    def polyy(self):
        """."""
        pvobj = self._pvs['polyy']
        if pvobj.connected:
            val = pvobj.value
            if val is not None:
                return val
        defv = _np.zeros(15, dtype=float)
        defv[0] = 1
        return defv

    @property
    def ksum(self):
        """."""
        defv = 1
        pvobj = self._pvs['ksum']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def arraya(self):
        """."""
        pvobj = self._pvs['arraya']
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayb(self):
        """."""
        pvobj = self._pvs['arrayb']
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayc(self):
        """."""
        pvobj = self._pvs['arrayc']
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayd(self):
        """."""
        pvobj = self._pvs['arrayd']
        return pvobj.get() if pvobj.connected else None

    @property
    def mtposx(self):
        """."""
        pvobj = self._pvs['arrayx']
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit * val

    @property
    def mtposy(self):
        """."""
        pvobj = self._pvs['arrayy']
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit * val

    @property
    def mtsum(self):
        """."""
        pvobj = self._pvs['arrays']
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return val

    @property
    def offsetx(self):
        """."""
        pvobj = self._pvs['offsetx']
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit * val

    @property
    def offsety(self):
        """."""
        pvobj = self._pvs['offsety']
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit * val

    @property
    def ctrl(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQCtrl']
        pvobj = self._config_pvs_rb["ACQTriggerEvent"]
        return pvobj.value if pvobj.connected else None

    @ctrl.setter
    def ctrl(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQCtrl']
        pvobj = self._config_pvs_sp["ACQTriggerEvent"]
        # self._config_ok_vals['ACQCtrl'] = val
        self._config_ok_vals["ACQTriggerEvent"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_type(self):
        """."""
        pvobj = self._config_pvs_rb["ACQChannel"]
        return pvobj.value if pvobj.connected else None

    @acq_type.setter
    def acq_type(self, val):
        """."""
        pvobj = self._config_pvs_sp["ACQChannel"]
        self._config_ok_vals["ACQChannel"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trigger(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQTriggerType']
        pvobj = self._config_pvs_rb["ACQTrigger"]
        return pvobj.value if pvobj.connected else None

    @acq_trigger.setter
    def acq_trigger(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQTriggerType']
        pvobj = self._config_pvs_sp["ACQTrigger"]
        # self._config_ok_vals['ACQTriggerType'] = val
        self._config_ok_vals["ACQTrigger"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_repeat(self):
        """."""
        pvobj = self._config_pvs_rb["ACQTriggerRep"]
        return pvobj.value if pvobj.connected else None

    @acq_repeat.setter
    def acq_repeat(self, val):
        """."""
        pvobj = self._config_pvs_sp["ACQTriggerRep"]
        self._config_ok_vals["ACQTriggerRep"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datatype(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQTriggerDataChan']
        pvobj = self._config_pvs_rb["ACQDataTrigChan"]
        return pvobj.value if pvobj.connected else None

    @acq_trig_datatype.setter
    def acq_trig_datatype(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQTriggerDataChan']
        pvobj = self._config_pvs_sp["ACQDataTrigChan"]
        # self._config_ok_vals['ACQTriggerDataChan'] = val
        self._config_ok_vals["ACQDataTrigChan"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["TbTPhaseSyncEn"]
        return pvobj.value if pvobj.connected else None

    @tbt_sync_enbl.setter
    def tbt_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["TbTPhaseSyncEn"]
        self._config_ok_vals["TbTPhaseSyncEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def fofb_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["FOFBPhaseSyncEn"]
        return pvobj.value if pvobj.connected else None

    @fofb_sync_enbl.setter
    def fofb_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["FOFBPhaseSyncEn"]
        self._config_ok_vals["FOFBPhaseSyncEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def sw_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["SwPhaseSyncEn"]
        return pvobj.value if pvobj.connected else None

    @sw_sync_enbl.setter
    def sw_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["SwPhaseSyncEn"]
        self._config_ok_vals["SwPhaseSyncEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def facq_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["FAcqPhaseSyncEn"]
        return pvobj.value if pvobj.connected else None

    @facq_sync_enbl.setter
    def facq_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["FAcqPhaseSyncEn"]
        self._config_ok_vals["FAcqPhaseSyncEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def monit_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["MonitPhaseSyncEn"]
        return pvobj.value if pvobj.connected else None

    @monit_sync_enbl.setter
    def monit_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["MonitPhaseSyncEn"]
        self._config_ok_vals["MonitPhaseSyncEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def test_data_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["TestDataEn"]
        return pvobj.value if pvobj.connected else None

    @test_data_enbl.setter
    def test_data_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["TestDataEn"]
        self._config_ok_vals["TestDataEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_enbl(self):
        """."""
        pvobj = self._config_pvs_rb["TbTDataMaskEn"]
        return pvobj.value if pvobj.connected else None

    @tbt_mask_enbl.setter
    def tbt_mask_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp["TbTDataMaskEn"]
        self._config_ok_vals["TbTDataMaskEn"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_begin(self):
        """."""
        pvobj = self._config_pvs_rb["TbTDataMaskSamplesBeg"]
        return pvobj.value if pvobj.connected else None

    @tbt_mask_begin.setter
    def tbt_mask_begin(self, val):
        """."""
        pvobj = self._config_pvs_sp["TbTDataMaskSamplesBeg"]
        self._config_ok_vals["TbTDataMaskSamplesBeg"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_end(self):
        """."""
        pvobj = self._config_pvs_rb["TbTDataMaskSamplesEnd"]
        return pvobj.value if pvobj.connected else None

    @tbt_mask_end.setter
    def tbt_mask_end(self, val):
        """."""
        pvobj = self._config_pvs_sp["TbTDataMaskSamplesEnd"]
        self._config_ok_vals["TbTDataMaskSamplesEnd"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def polycal(self):
        """."""
        pvobj = self._config_pvs_rb["XYPosCal"]
        return pvobj.value if pvobj.connected else None

    @polycal.setter
    def polycal(self, val):
        """."""
        val = _CSBPM.DsblEnbl.enabled if val else _CSBPM.DsblEnbl.disabled
        pv1 = self._config_pvs_sp["XYPosCal"]
        pv2 = self._config_pvs_sp["SumPosCal"]
        self._config_ok_vals["XYPosCal"] = val
        self._config_ok_vals["SumPosCal"] = val
        if self.put_enable and pv1.connected:
            pv1.put(val, wait=False)
        if self.put_enable and pv2.connected:
            pv2.put(val, wait=False)

    @property
    def nrsamplespost(self):
        """."""
        pvobj = self._config_pvs_rb["ACQSamplesPost"]
        return pvobj.value if pvobj.connected else None

    @nrsamplespost.setter
    def nrsamplespost(self, val):
        """."""
        pvobj = self._config_pvs_sp["ACQSamplesPost"]
        self._config_ok_vals["ACQSamplesPost"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def nrsamplespre(self):
        """."""
        pvobj = self._config_pvs_rb["ACQSamplesPre"]
        return pvobj.value if pvobj.connected else None

    @nrsamplespre.setter
    def nrsamplespre(self, val):
        """."""
        pvobj = self._config_pvs_sp["ACQSamplesPre"]
        self._config_ok_vals["ACQSamplesPre"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def nrshots(self):
        """."""
        pvobj = self._config_pvs_rb["ACQShots"]
        return pvobj.value if pvobj.connected else None

    @nrshots.setter
    def nrshots(self, val):
        """."""
        pvobj = self._config_pvs_sp["ACQShots"]
        self._config_ok_vals["ACQShots"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    def calc_sp_multiturn_pos(self, **kwargs):
        """."""
        nturns = kwargs.get("nturns", 1)
        refx = kwargs.get("refx", 0.0)
        refy = kwargs.get("refy", 0.0)
        refsum = kwargs.get("refsum", 0.0)
        maskbeg = kwargs.get("maskbeg", 0)
        maskend = kwargs.get("maskend", 0)

        wsize = self.tbtrate
        maskbeg = min(maskbeg, wsize - 2)
        maskend = min(maskend, wsize - maskbeg - 2)
        mask = slice(maskbeg, wsize - maskend)

        # NOTE: I have to invert array B with C here because of the way
        # the ADCSWAP rate works. Fixed in 2020/07/01 after talking to
        # Daniel Tavares.
        vals = {
            "A": self.arraya,
            "C": self.arrayb,
            "B": self.arrayc,
            "D": self.arrayd,
        }
        siz = None
        for _, val in vals.items():
            if val is None or val.size == 0:
                siz = 0
                break
            nzrs = val.size
            siz = nzrs if siz is None else min(siz, nzrs)

        x_cal = _np.full(nturns, refx)
        y_cal = _np.full(nturns, refy)
        s_cal = _np.full(nturns, refsum)

        # handle cases where length read is smaller than required.
        rnts = min(siz // wsize, nturns)
        if not (siz and rnts):
            return x_cal, y_cal, s_cal

        for key, val in vals.items():
            val = val[: (rnts * wsize)]
            val = val.reshape(-1, wsize)[:, mask]
            vals[key] = _np.std(val, axis=1)

        sum1, sum2 = vals["A"] + vals["C"], vals["D"] + vals["B"]
        not_zero = _np.logical_not(_np.isclose(sum1, 0.0))
        not_zero &= _np.logical_not(_np.isclose(sum2, 0.0))
        diff1 = (vals["A"][not_zero] - vals["C"][not_zero]) / sum1[not_zero]
        diff2 = (vals["D"][not_zero] - vals["B"][not_zero]) / sum2[not_zero]
        x_raw = (diff1 + diff2) / 2
        y_raw = (diff1 - diff2) / 2
        if self._config_ok_vals["XYPosCal"] == _CSBPM.DsblEnbl.enabled:
            x_raw, y_raw = self._apply_polyxy(x_raw, y_raw)

        x_cal[:rnts][not_zero] = x_raw * self.poskx
        y_cal[:rnts][not_zero] = y_raw * self.posky
        x_cal[:rnts][not_zero] *= self._orb_conv_unit
        y_cal[:rnts][not_zero] *= self._orb_conv_unit
        x_cal[:rnts][not_zero] -= self.offsetx or 0.0
        y_cal[:rnts][not_zero] -= self.offsety or 0.0
        s_cal[:rnts] = (sum1 + sum2) * self.ksum
        return x_cal, y_cal, s_cal

    def _apply_polyxy(self, x_raw, y_raw):
        """."""
        x_pol = self._calc_poly(x_raw, y_raw, plane="x")
        y_pol = self._calc_poly(y_raw, x_raw, plane="y")
        return x_pol, y_pol

    def _calc_poly(self, th1, ot1, plane="x"):
        """."""
        ot2 = ot1 * ot1
        ot4 = ot2 * ot2
        ot6 = ot4 * ot2
        ot8 = ot4 * ot4
        th2 = th1 * th1
        th3 = th2 * th1
        th5 = th3 * th2
        th7 = th5 * th2
        th9 = th7 * th2
        pol = self.polyx if plane == "x" else self.polyy

        return (
            th1
            * (
                pol[0]
                + ot2 * pol[1]
                + ot4 * pol[2]
                + ot6 * pol[3]
                + ot8 * pol[4]
            )
            + th3 * (pol[5] + ot2 * pol[6] + ot4 * pol[7] + ot6 * pol[8])
            + th5 * (pol[9] + ot2 * pol[10] + ot4 * pol[11])
            + th7 * (pol[12] + ot2 * pol[13])
            + th9 * pol[14]
        )

    def _reset_has_news(self, *args, **kwargs):
        _ = args, kwargs
        self.has_news = True


class TimingConfig(_BaseTimingConfig):
    """."""

    def __init__(self, acc, callback=None):
        """."""
        super().__init__(acc, callback=callback)
        trig = self._csorb.trigger_acq_name
        evg = self._csorb.evg_name
        opt = {"connection_timeout": TIMEOUT}
        self._config_ok_vals = {
            "NrPulses": 1,
            "State": _TIConst.TrigStates.Enbl,
        }
        if _HLTimesearch.has_delay_type(trig):
            self._config_ok_vals["RFDelayType"] = _TIConst.TrigDlyTyp.Manual
        pref_name = LL_PREF + trig + ":"
        self._config_pvs_rb = {
            "Delay": _PV(pref_name + "Delay-RB", **opt),
            "TotalDelay": _PV(pref_name + "TotalDelay-Mon", **opt),
            "NrPulses": _PV(pref_name + "NrPulses-RB", **opt),
            "Duration": _PV(pref_name + "Duration-RB", **opt),
            "State": _PV(pref_name + "State-Sts", **opt),
            "Injecting": _PV(LL_PREF + evg + ":InjectionEvt-Sts", **opt),
            "EGTrig": _PV("LI-01:EG-TriggerPS:status", **opt),
        }
        self._config_pvs_sp = {
            "NrPulses": _PV(pref_name + "NrPulses-SP", **opt),
            "State": _PV(pref_name + "State-Sel", **opt),
        }
        if _HLTimesearch.has_delay_type(trig):
            self._config_pvs_rb["RFDelayType"] = _PV(
                pref_name + "RFDelayType-Sts", **opt
            )
            self._config_pvs_sp["RFDelayType"] = _PV(
                pref_name + "RFDelayType-Sel", **opt
            )

    @property
    def injecting(self):
        """."""
        inj = bool(self._config_pvs_rb["Injecting"].value)
        eg_trig = bool(self._config_pvs_rb["EGTrig"].value)
        return inj and eg_trig

    @property
    def nrpulses(self):
        """."""
        pvobj = self._config_pvs_rb["NrPulses"]
        return pvobj.value if pvobj.connected else None

    @nrpulses.setter
    def nrpulses(self, val):
        """."""
        pvobj = self._config_pvs_sp["NrPulses"]
        self._config_ok_vals["NrPulses"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def duration(self):
        """."""
        pvobj = self._config_pvs_rb["Duration"]
        return pvobj.value if pvobj.connected else None

    @property
    def delay(self):
        """."""
        pvobj = self._config_pvs_rb["Delay"]
        return pvobj.value if pvobj.connected else None

    @property
    def totaldelay(self):
        """."""
        pvobj = self._config_pvs_rb["TotalDelay"]
        return pvobj.value if pvobj.connected else None



"""ID feedforward App."""

import logging as _log
import os as _os
import time as _time
from copy import deepcopy as _dcopy

import epics as _epics
import numpy as _np

from ..callbacks import Callback as _Callback
from ..clientconfigdb import ConfigDBException as _ConfigDBException
from ..devices import IDFF as _IDFF
from ..util import update_bit as _updt_bit
from .csdev import ETypes as _ETypes, IDFFConst as _Const


class App(_Callback):
    """Main application for ID feedforward."""

    DEF_CORR_RAMP_NRPTS = 20
    DEF_PS_TIMEOUT = 5  # [s]
    DEF_MAX_DCURR = 0.1  # [A]

    def __init__(self, idname,
                 enbl_chcorrs, enbl_cvcorrs,
                 enbl_qscorrs, enbl_lccorrs,
                 enbl_qncorrs, enbl_cccorrs):
        """Class constructor."""
        super().__init__()
        self.const = _Const(
            idname,
            enbl_chcorrs, enbl_cvcorrs,
            enbl_qscorrs, enbl_lccorrs,
            enbl_qncorrs, enbl_cccorrs)
        self.pvs_prefix = self.const.idffname
        self.pvs_database = self.const.get_propty_database()

        self.loop_state = _Const.DEFAULT_LOOP_STATE
        self.loop_freq = _Const.DEFAULT_LOOP_FREQ
        self.control_ch = self.const.enbl_chcorrs
        self.control_cv = self.const.enbl_cvcorrs
        self.control_qs = self.const.enbl_qscorrs
        self.control_lc = self.const.enbl_lccorrs
        self.control_qn = self.const.enbl_qncorrs
        self.control_cc = self.const.enbl_cccorrs
        self.polarization = 'none'
        self.corr_ramp_status = _Const.DEFAULT_CORR_RAMP_STATUS

        # IDFF object with IDFF config
        idffname_soft = self.const.idffname
        self.idff = _IDFF(idffname_soft, with_devctrl=False)

        self.config_name = ''  # stored in autosave file
        self.read_autosave_file()

        # load idff in configdb
        self.load_config(self.config_name)

        # pvs to write methods
        self.map_pv2write = {
            'LoopState-Sel': self.write_loop_state,
            'LoopFreq-SP': self.write_loop_freq,
            'ConfigName-SP': self.write_config_name,
            'CorrConfig-Cmd': self.write_cmd_corr_config,
            'CorrSaveOffsets-Cmd': self.write_cmd_save_offsets,
            'CorrRamp-Cmd': self.write_cmd_corr_ramp,
            # following mappings are only used if
            # corresponding correctors are in IDFF
            'ControlCH-Sel': self.write_control_ch,
            'ControlCV-Sel': self.write_control_cv,
            'ControlQS-Sel': self.write_control_qs,
            'ControlLC-Sel': self.write_control_lc,
            'ControlQN-Sel': self.write_control_qn,
            'ControlCC-Sel': self.write_control_cc,
        }

        self._thread_quit = False
        self._corr_setpoints = None
        self._thread_ff = _epics.ca.CAThread(
            target=self.feedforward_loop, daemon=True)
        self._thread_ff.start()

    def init_database(self):
        """Set initial PV values."""
        pvn2vals = {
            'LoopState-Sel': self.loop_state,
            'LoopState-Sts': self.loop_state,
            'LoopFreq-SP': self.loop_freq,
            'LoopFreq-RB': self.loop_freq,
            'ConfigName-SP': self.config_name,
            'ConfigName-RB': self.config_name,
            'Polarization-Mon': self.polarization,
            'CorrConfig-Cmd': 0,
            'CorrStatus-Mon': _Const.DEFAULT_CORR_STATUS,
        }
        if self.const.enbl_chcorrs:
            pvn2vals.update({
                'ControlCH-Sel': self.control_ch,
                'ControlCH-Sts': self.control_ch,
                })
        if self.const.enbl_cvcorrs:
            pvn2vals.update({
                'ControlCV-Sel': self.control_cv,
                'ControlCV-Sts': self.control_cv,
                })
        if self.const.enbl_qscorrs:
            pvn2vals.update({
                'ControlQS-Sel': self.control_qs,
                'ControlQS-Sts': self.control_qs,
                })
        if self.const.enbl_lccorrs:
            pvn2vals.update({
                'ControlLC-Sel': self.control_lc,
                'ControlLC-Sts': self.control_lc,
                })
        if self.const.enbl_qncorrs:
            pvn2vals.update({
                'ControlQN-Sel': self.control_qn,
                'ControlQN-Sts': self.control_qn,
                })
        if self.const.enbl_cccorrs:
            pvn2vals.update({
                'ControlCC-Sel': self.control_cc,
                'ControlCC-Sts': self.control_cc,
                })
        for pvn, val in pvn2vals.items():
            self.run_callbacks(pvn, val)
        self.update_log('Started.')

    def process(self, interval):
        """Sleep."""
        # check correctors state periodically
        _t0 = _time.time()
        self.update_corr_status()
        self.update_corr_pvs()
        dtime = _time.time() - _t0
        sleep_time = interval - dtime
        # sleep
        if sleep_time > 0:
            _time.sleep(sleep_time)
        else:
            strf = f'process took {(dtime)*1000:f}ms.'
            _log.debug(strf)

    def read(self, reason):
        """Read from IOC database."""
        _ = reason
        value = None
        return value

    # ----- write methods -----

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
            _log.info('%s Write for: %s --> %s',
                      str(status).upper(), reason, str(value))
            return status
        _log.warning('PV %s does not have a set function.', reason)
        return False

    def write_loop_state(self, value):
        """Set loop state."""
        if not 0 <= value < len(_ETypes.OPEN_CLOSED):
            return False
        self.loop_state = value
        act = ('En' if value else 'Dis')
        self.update_log(f'Feedforward loop {act}abled.')
        self.run_callbacks('LoopState-Sts', value)
        return True

    def write_loop_freq(self, value):
        """Set loop frequency."""
        fmin, fmax = _Const.DEFAULT_LOOP_FREQ_MIN, _Const.DEFAULT_LOOP_FREQ_MAX
        if not fmin <= value < fmax:
            return False
        self.loop_freq = value
        self.update_log(f'Loop frequency updated to {value:.3f}Hz.')
        self.run_callbacks('LoopFreq-RB', value)
        return True

    def write_control_ch(self, value):
        """Set whether to include CH or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_ch = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled CH control.')
        self.run_callbacks('ControlCH-Sts', value)
        return True

    def write_control_cv(self, value):
        """Set whether to include CV or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_cv = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled CV control.')
        self.run_callbacks('ControlCV-Sts', value)
        return True

    def write_control_qs(self, value):
        """Set whether to include QS or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_qs = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled QS control.')
        self.run_callbacks('ControlQS-Sts', value)
        return True

    def write_control_lc(self, value):
        """Set whether to include LC or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_lc = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled LC control.')
        self.run_callbacks('ControlLC-Sts', value)
        return True

    def write_control_qn(self, value):
        """Set whether to include QN or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_qn = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled QN control.')
        self.run_callbacks('ControlQN-Sts', value)
        return True

    def write_control_cc(self, value):
        """Set whether to include CC or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self.control_cc = value
        act = ('En' if value else 'Dis')
        self.update_log(f'{act}abled CC control.')
        self.run_callbacks('ControlCC-Sts', value)
        return True

    def write_config_name(self, value, save_autoconfig=True):
        """Set configuration name."""
        if self.loop_state == self.const.LoopState.Closed:
            self.update_log('ERR:Open loop before changing configuration.')
            return False

        if not self.load_config(value):
            return False
        self.config_name = value
        self.run_callbacks('ConfigName-RB', value)

        if save_autoconfig:
            self.update_autosave_file()

        return True

    def write_cmd_corr_config(self, _):
        """Command to reconfigure power supplies to desired state."""
        if self.loop_state == self.const.LoopState.Closed:
            self.update_log('ERR:Open loop before configure correctors.')
            return False

        corrdevs = self.get_control_corrdevs()
        for dev in corrdevs:
            # turn on
            if not dev.cmd_turn_on():
                return False
            # opmode slowref
            if not dev.cmd_slowref():
                return False

        return True

    def write_cmd_corr_ramp(self, _):
        """Command to implement power supplies current ramp."""
        # initial checks
        ramp_on_status = _ETypes.CORR_RAMP_STATUS.index('On')
        ramp_off_status = _ETypes.CORR_RAMP_STATUS.index('Off')
        if self.loop_state or self.corr_ramp_status == ramp_on_status:
            # return if FF loop is closed ot corr ramp is being executed
            return True
        else:
            self.corr_ramp_status = ramp_on_status
            self.run_callbacks(
                self.pvs_prefix + ':CorrRampStatus-Mon', ramp_on_status)

        # update polarization state
        self.update_polarization()

        # correctors value calculation
        self.update_corr_setpoints()
        corr_setpoints_init = _dcopy(self._corr_setpoints)
        if not corr_setpoints_init:
            self.corr_ramp_status = ramp_off_status
            self.run_callbacks(
                self.pvs_prefix + ':CorrRampStatus-Mon', ramp_off_status
            )
            return False

        # current and calculated setpoints
        sp_now = self.idff.read_setpoints()
        sp_cal, *_ = corr_setpoints_init

        # calc nrpts in ramp
        max_dcurr = 0
        for pvname, value in sp_now.items():
            max_dcurr = max(max_dcurr, abs(value - sp_cal[pvname]))
        nrpts = int(max_dcurr / self.DEF_MAX_DCURR) + 1

        # build ramp
        ramp = dict()
        for pvname, value in sp_now.items():
            rmp = _np.linspace(value, sp_cal[pvname], nrpts)
            ramp[pvname] = rmp[1:]

        # implement ramp
        for idx in range(nrpts):
            # prepare corr setpoints
            for pvname, value in ramp.items():
                if self._corr_setpoints:
                    self._corr_setpoints[0][pvname] = value[idx]
            # implement setpoints
            self.implement_setpoints()
            # sleep
            _time.sleep(0.05)

        # finalize
        self.corr_ramp_status = ramp_off_status
        self.run_callbacks(
                self.pvs_prefix + ':CorrRampStatus-Mon', ramp_off_status)

    def write_cmd_save_offsets(self, _):
        """Save current values of correctors' currents to si_idff configdb."""
        if self.loop_state == self.const.LoopState.Closed:
            self.update_log('ERR:Open loop before saving correctors offsets.')
            return False

        return self.save_config_offsets()

    # ----- feedforward loop methods -----

    @property
    def thread_is_alive(self):
        """Return if thread is alive."""
        return self._thread_ff.is_alive()

    @property
    def thread_quit(self):
        """Quit and shutdown threads."""
        return self._thread_quit

    @thread_quit.setter
    def thread_quit(self, value):
        if value:
            self._thread_quit = value

    def feedforward_loop(self):
        """Main IDFF loop."""
        self.idff.wait_for_connection(timeout=2)

        while not self._thread_quit:
            # updating interval
            tplanned = 1.0/self.loop_freq

            # initial time
            _t0 = _time.time()

            # check IDFF device connection
            if not self.idff.connected:
                self.update_log('WARN: IDFF device disconnected!')
                self._do_sleep(_t0, tplanned)
                continue

            # update polarization state
            self.update_polarization()

            # correctors value calculation
            self.update_corr_setpoints()

            # return if loop is not closed
            if not self.loop_state:
                self._do_sleep(_t0, tplanned)
                continue

            # setpoints implementation
            if self._corr_setpoints:
                self.implement_setpoints()

            # sleep unused time or signal overtime to stdout
            self._do_sleep(_t0, tplanned)
        self._thread_quit = False

    def implement_setpoints(self):
        """."""
        corrdevs = self.get_control_corrdevs()

        try:
            # use PS IOCs (IDFF) to implement setpoints
            setpoints, *_ = self._corr_setpoints  # type: ignore
            self.idff.implement_setpoints(
                setpoints=setpoints, corrdevs=corrdevs)

        except ValueError as err:
            self.update_log('ERR:'+str(err))

    # ----- read and save methods -----

    def read_autosave_file(self):
        """Read autosave file."""
        # TODO: create a autosave/restore mechanism that is
        # standardized for all IOCS!
        filename = self.const.autosave_fname
        config_name = None
        if _os.path.isfile(filename):
            with open(filename, 'r') as fil:
                config_name = fil.read().strip('\n')
        if config_name is not None:
            self.write_config_name(config_name, save_autoconfig=False)
        else:
            _log.info(
                'No backup file was found, default value '
                'was used for si_idff configuration name.')
            self.config_name = self.const.configname + '_ref'
            self.update_autosave_file()
            self.update_log('First autosave file was created.')

    def update_autosave_file(self):
        """Update autosave file."""
        path = _os.path.split(self.const.autosave_fname)[0]
        _os.makedirs(path, exist_ok=True)
        with open(self.const.autosave_fname, 'w+') as fil:
            fil.write(self.config_name)

    def load_config(self, config_name):
        """."""
        try:
            self.idff.load_config(config_name)
            self.update_log(f'Updated configuration: {config_name}.')
        except (ValueError, _ConfigDBException) as err:
            self.update_log('ERR: could not load config ' + self.config_name)
            raise err
        return True

    def save_config_offsets(self):
        """."""
        try:
            offsets = self.idff.read_corr_offset_values()
        except:  # noqa: E722
            self.update_log('ERR: could not read correctors offsets')
            return False
        if not offsets:
            self.update_log('WARN: Offsets not defined for this IDFF')
            return False

        config = self.idff.idffconfig
        value = config.value
        value['offsets'] = offsets
        try:
            config.save()
        except:  # noqa: E722
            self.update_log('ERR: could not save configuration')
            return False
        self.update_log('Correctors offsets saved in ' + self.config_name)

    # ----- update methods -----

    def update_polarization(self):
        """."""
        new_pol, *_ = self.idff.get_polarization_state()
        if new_pol is not None and new_pol != self.polarization:
            self.polarization = new_pol
            self.run_callbacks('Polarization-Mon', new_pol)

    def update_corr_setpoints(self):
        """."""
        try:
            self._corr_setpoints = self.idff.calculate_setpoints()
            # setpoints, polarization, *parameters = self._corr_setpoints
            # pparameter_value, kparameter_value = parameters
            # print('pparameter: ', pparameter_value)
            # print('kparameter: ', kparameter_value)
            # print('polarization: ', polarization)
            # print('setpoints: ', setpoints)
            # print()
        except ValueError as err:
            self.update_log('ERR:'+str(err))

    def update_corr_status(self):
        """Update CorrStatus-Mon PV."""
        corrdevs = self.get_control_corrdevs()

        status = 0
        if all(d.connected for d in corrdevs):
            if any(d.pwrstate != d.PWRSTATE.On for d in corrdevs):
                status = _updt_bit(status, 1, 1)
            if any(d.opmode != d.OPMODE_STS.SlowRef for d in corrdevs):
                status = _updt_bit(status, 2, 1)
        else:
            status = _Const.DEFAULT_CORR_STATUS

        self.run_callbacks('CorrStatus-Mon', status)

    def update_corr_pvs(self):
        """Update corrector setpoint PVs."""
        if self._corr_setpoints is None:
            return
        setpoints, *_ = self._corr_setpoints
        idff = self.idff
        corrnames = idff.chnames + idff.cvnames + \
            idff.qsnames + idff.lcnames + idff.qnnames + \
            idff.ccnames
        corrlabels = (
            'CH_1', 'CH_2', 'CV_1', 'CV_2',
            'QS_1', 'QS_2', 'LCH', 'LCV',
            'QD1_1', 'QF_1', 'QD2_1', 'QD2_2', 'QF_2', 'QD1_2',
            'CC1_1', 'CC2_1', 'CC2_2', 'CC1_2',
            )
        for corrlabel, corrname in zip(corrlabels, corrnames):
            for corr_pvname in setpoints:
                if corrname in corr_pvname:
                    pvname = 'Corr' + corrlabel + 'Current-Mon'
                    value = setpoints[corr_pvname]
                    if pvname in self.pvs_database:
                        self.run_callbacks(pvname, value)

    def update_log(self, msg):
        """."""
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    #  ----- auxiliary methods -----

    def get_control_corrdevs(self):
        """Return devices of currently enabled correctors."""
        corrdevs = list()
        if self.control_ch == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.chdevs)
        if self.control_cv == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.cvdevs)
        if self.control_qs == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.qsdevs)
        if self.control_lc == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.lcdevs)
        if self.control_qn == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.qndevs)
        if self.control_cc == self.const.DsblEnbl.Enbl:
            corrdevs.extend(self.idff.ccdevs)
        return corrdevs

    # ----- update pvs methods -----

    def _do_sleep(self, time0, tplanned):
        ttook = _time.time() - time0
        tsleep = tplanned - ttook
        if tsleep > 0:
            _time.sleep(tsleep)
        else:
            strf = (
                f'Feedforward step took more than planned... '
                f'{ttook:.3f}/{tplanned:.3f} s')
            _log.warning(strf)



"""SI fast orbit correction power supply matrix configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template),
        'check': False,
    }
    return _dict


#  Fast Orbit Correction Power Supply Matrix for Sirius:
#    Each row refers to a fast corrector
#   -First column is the Kp for each power supply
#   -Second column is the Ti for each power supply
#   -Third column is the Acc filter gain for each power supply
#   -From the fourth to the last nth column,
#    we have the acc filter coefficients
#   -By standard, we consider 4 biquads for each corrector.
#    One biquad has 5 coefficients
#
# | KPi  TIi  GAINi  COEFFi        COEFFn |
# |  ................................... |
# | KPj  TIj  GAINj  COEFFj       COEFFn |

_template = 160*[23*[0.0]]



"""Main Module of the IOC Logic."""

import logging as _log
import time as _time
from copy import deepcopy as _dcopy
from threading import Thread as _Thread

import numpy as _np
import scipy.optimize as _scyopt
import scipy.signal as _scysig

from ..callbacks import Callback as _Callback
from ..envars import VACA_PREFIX as _VACA_PREFIX
from ..epics import PV as _PV
from ..oscilloscope import Keysight as _Keysight, ScopeSignals as _ScopeSignals
from .csdev import Const as _Const, get_si_fpmosc_database as _get_database


class FPMOscApp(_Callback):
    """Main Class."""
    Const = _Const

    def __init__(self):
        """Class constructor."""
        super().__init__()
        self._pvs_database = _get_database()
        self._prefix = _VACA_PREFIX + ('-' if _VACA_PREFIX else '')

        # initialize vars
        self._time0 = _time.time()
        self._fillpat_fid_offset = 0
        self._fillpat_update_time = 5  # [s]
        self._fillpat_ref = _np.ones(_Const.FP_HARM_NUM) / _Const.FP_HARM_NUM
        self._fillpat_osc = _Keysight(
            scopesignal=_ScopeSignals.SI_FILL_PATTERN
        )
        self._fillpat_thread = None

        # pvs
        self._current_pv = _PV(
            self._prefix+'SI-Glob:AP-CurrInfo:Current-Mon',
            connection_timeout=0.05
        )
        self._rffreq_pv = _PV(
            self._prefix+'RF-Gen:GeneralFreq-RB', connection_timeout=0.05
        )

        self._current_pv.add_callback(self._callback_get_current)

    @property
    def pvs_database(self):
        """."""
        return _dcopy(self._pvs_database)

    def process(self, interval):
        """."""
        _time.sleep(interval)

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks('UpdateTime-SP', self._fillpat_update_time)
        self.run_callbacks('UpdateTime-RB', self._fillpat_update_time)
        self.run_callbacks('FiducialOffset-SP', self._fillpat_fid_offset)
        self.run_callbacks('FiducialOffset-RB', self._fillpat_fid_offset)
        self.run_callbacks('FillPatternRef-SP', self._fillpat_ref)
        self.run_callbacks('FillPatternRef-RB', self._fillpat_ref)

    def read(self, reason):
        """Read from IOC database."""
        _ = reason
        return None

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        if reason == 'UpdateTime-SP':
            self._fillpat_update_time = min(
                _Const.FP_MAX_UPDT_TIME, max(0, float(value))
            )
            self.run_callbacks('UpdateTime-RB', self._fillpat_update_time)
            return True
        elif reason == 'FiducialOffset-SP':
            self._fillpat_fid_offset = min(
                _Const.FP_HARM_NUM, max(-_Const.FP_HARM_NUM, int(value))
            )
            self.run_callbacks('FiducialOffset-RB', self._fillpat_fid_offset)
            return True
        elif reason == 'FillPatternRef-SP':
            if not isinstance(value, (_np.ndarray, list, tuple)):
                return False
            value = _np.array(value)
            if value.size != 864 or value.min() < 0:
                return False
            self._fillpat_ref = value / value.sum()
            self.run_callbacks('FillPatternRef-RB', self._fillpat_ref)
            return True
        return False

    def _callback_get_current(self, pvname, value, **kws):
        _ = kws, pvname

        # trigger update of filling pattern from oscilloscope.
        if (
            self._fillpat_thread is None or
            not self._fillpat_thread.is_alive()
        ):
            self._fillpat_thread = _Thread(
                target=self._update_filling_pattern,
                args=(value, ),
                daemon=True,
            )
            self._fillpat_thread.start()

    def _update_filling_pattern(self, current):
        t0_ = _time.time()
        frf = 499667557.37  # default value
        if self._rffreq_pv.connected:
            frf = self._rffreq_pv.value
        # first peak in Hilbert transform is compromized by dft issues, so get
        # data starting from second:
        bun_spacing = _np.arange(1, _Const.FP_HARM_NUM + 1) / frf * 1e9  # [ns]

        try:
            tim, fill = self._fillpat_osc.wfm_get_data()
            tim = tim[:_Const.FP_MAX_ARR_SIZE]
            fill = fill[:_Const.FP_MAX_ARR_SIZE]
        except Exception:
            _log.warning('Could not read oscilloscope data.')
            return
        tim *= 1e9  # from [s] to [ns]
        tim -= tim[0]
        fill -= fill.mean()
        # NOTE: I tried to filter the raw data at the harmonics of the
        # revolution frequency, around the RF frequency, but it didn't improve
        # the results. I decided to keep the processing as simple as possible.
        hil = _np.abs(_scysig.hilbert(fill))

        if tim[-1] < bun_spacing[-1]:
            _log.warning('oscilloscope data is incomplete.')
            return

        def get_interp(off, return_res=False):
            tim2ns = bun_spacing + off
            fil2ns = _np.interp(tim2ns, tim, hil)

            if return_res:
                return tim2ns, fil2ns
            hil_max = hil[200:-200].max()  # avoid borders
            weight = fil2ns[fil2ns > hil_max * 20/100]
            obj = weight - hil_max
            return obj * _np.sqrt(weight)

        try:
            init_guess_offset = 0.2  # [ns]
            res = _scyopt.least_squares(get_interp, init_guess_offset)
        except Exception:
            _log.warning('Fitting did not work.')
            return
        tim2ns, fil2ns = get_interp(res.x, return_res=True)
        fid = self._fillpat_fid_offset
        fil2ns = _np.roll(fil2ns, -fid)
        tim2ns = _np.roll(tim2ns, -fid)

        # Scale filling pattern so that its sum is equal to
        # the total current stored
        if current is None:
            _log.warning('Could not read current from DCCT.')
            return
        elif current <= 0.0:
            return
        fac = current / fil2ns.sum()
        fil2ns *= fac
        fill *= fac
        hil *= fac

        # Equivalent current for broad band induced heating. See:
        # https://accelconf.web.cern.ch/ipac2024/pdf/THPC44.pdf
        # for details
        equiv_curr = _np.sqrt(_np.sum(fil2ns**2)*_Const.FP_HARM_NUM)  # [mA]

        avg_curr = current / _Const.FP_HARM_NUM
        filref = self._fillpat_ref * current
        fil_err = _np.std(filref - fil2ns) / avg_curr * 100
        # Compute KL Divergence between distributions:
        idx = self._fillpat_ref > 0
        fil_kld = _np.sum(
            self._fillpat_ref[idx] * _np.log2(filref[idx]/fil2ns[idx]))

        # Update PVs:
        self.run_callbacks('UniFillEqCurrent-Mon', equiv_curr)
        self.run_callbacks('ErrorRelStd-Mon', fil_err)
        self.run_callbacks('ErrorKLDiv-Mon', fil_kld)
        self.run_callbacks('FillPattern-Mon', fil2ns)
        self.run_callbacks('FillPatternRef-Mon', filref)
        self.run_callbacks('Time-Mon', tim2ns)
        self.run_callbacks('TimeOffset-Mon', res.x)
        self.run_callbacks('Raw-Mon', fill)
        self.run_callbacks('RawAmp-Mon', hil)
        self.run_callbacks('RawTime-Mon', tim)

        # This thread is triggered by the update of the current, which is
        # faster than the frequency we want to update the filling pattern.
        # Hence, we need to wait here:
        dt_ = self._fillpat_update_time - (_time.time() - t0_)
        _time.sleep(max(dt_, 0))



"""Beamline devices."""

from .device import Device as _Device


class CAXCtrl(_Device):
    """Carcara beamline control.

    * Beamline commissionning:
    https://cnpemcamp.sharepoint.com/:p:/s/lnls/groups/opt/EZB-ePPBFvRPqi5IM1fjTsoB7uwtUhUrw3inlGnRfDm1KA?e=4lP3Bi&clickparams=eyJBcHBOYW1lIjoiVGVhbXMtRGVza3RvcCIsIkFwcFZlcnNpb24iOiIxNDE1LzI0MDYyNzI0ODE0In0%3D

    * DVF:
    https://cnpemcamp.sharepoint.com/:p:/s/lnls/groups/opt/EfXWP0gm-URPt1mzEZEWxbABzJxfJg3kt86jOJrV3KQoMg?e=eaT4T1&clickparams=eyJBcHBOYW1lIjoiVGVhbXMtRGVza3RvcCIsIkFwcFZlcnNpb24iOiIxNDE1LzI0MDYyNzI0ODE0In0%3D

    * M1 online autodesk drawing:
    https://drive.autodesk.com/de29ccb7d/shares/SH9285eQTcf875d3c5396deb53e725438482
    """

    _DEFAULT_MOTOR_TIMEOUT = 2.0  # [s]

    PROPERTIES_DEFAULT = (
        # mirror controls
        'A:PP01:A',
        'A:PP01:A.RBV',
        'A:PP01:A.STOP',  # RotY
        'A:PP01:B',
        'A:PP01:B.RBV',
        'A:PP01:B.STOP',  # Tx
        'A:PP01:C',
        'A:PP01:C.RBV',
        'A:PP01:C.STOP',  # Y1
        'A:PP01:D',
        'A:PP01:D.RBV',
        'A:PP01:D.STOP',  # Y2
        'A:PP01:E',
        'A:PP01:E.RBV',
        'A:PP01:E.STOP',  # Y3
        # mirror photocollector
        'A:RIO01:9215A:ai0',
        # mirror temperatures
        'A:RIO01:M1_CtrltempSp',
        'A:RIO01:9226B:temp0',
        'A:RIO01:9226B:temp1',
        'A:RIO01:9226B:temp2',
        'A:RIO01:9226B:temp3',
        'A:RIO01:9226B:temp4',
        # dvf motors
        'B:PP01:E',
        'B:PP01:E.RBV',
        'B:PP01:E.STOP',  # z pos
        'B:PP01:F',
        'B:PP01:F.RBV',
        'B:PP01:F.STOP',  # lens
    )

    def __init__(self, devname=None, props2init='all', **kwargs):
        """Init."""
        devname = 'CAX' if devname is None else devname
        super().__init__(devname, props2init=props2init, **kwargs)

    @property
    def m1_roty_pos(self):
        """Return the linear actuator pos related to RotY rotation [mm].

        RotY is performed by a linear actuator in one of
        longitudinal ends of the mirror. The mirror is pivoted
        at its longitudinal center and the linear actuator induces
        a rotation around the Y axis.
        """
        return self['A:PP01:A.RBV']

    @m1_roty_pos.setter
    def m1_roty_pos(self, value):
        """Set the linear actuator pos related to RotY rotation [mm].

        RotY is performed by a linear actuator in one of
        longitudinal ends of the mirror. The mirror is pivoted
        at its longitudinal center and the linear actuator induces
        a rotation around the Y axis.
        """
        self['A:PP01:A'] = value

    @property
    def m1_tx_pos(self):
        """Return the linear actuator pos related to Tx translation [mm].

        This linear actuator translates directly to the horizontal
        transverse position of the mirror.
        """
        return self['A:PP01:B.RBV']

    @m1_tx_pos.setter
    def m1_tx_pos(self, value):
        """Set the linear actuator pos related to Tx translation [mm].

        This linear actuator is directly related to the horizontal
        transverse position of the mirror.
        """
        self['A:PP01:B'] = value

    @property
    def m1_y1_pos(self):
        """Return the first linear vertical actuator pos Y1 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in oposite horizontal ends.
        """
        return self['A:PP01:C.RBV']

    @m1_y1_pos.setter
    def m1_y1_pos(self, value):
        """Set the first linear vertical actuator pos Y1 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in opposite horizontal ends.
        """
        self['A:PP01:C'] = value

    @property
    def m1_y2_pos(self):
        """Return the second linear vertical actuator pos Y2 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in opposite horizontal ends.
        """
        return self['A:PP01:D.RBV']

    @m1_y2_pos.setter
    def m1_y2_pos(self, value):
        """Set the second linear vertical actuator pos Y1 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in opposite horizontal ends.
        """
        self['A:PP01:D'] = value

    @property
    def m1_y3_pos(self):
        """Return the third linear actuator pos Y2 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in opposite horizontal ends.
        """
        return self['A:PP01:E.RBV']

    @m1_y3_pos.setter
    def m1_y3_pos(self, value):
        """Set the third linear vertical actuator pos Y1 [mm].

        Rotations RotX, RotZ and translation Ty are implemented as combinations
        of three vertical independent actuators. Y1 actuator is located in one
        longitudinal side of the mirror base whereas Y2 amd Y3 are located in
        the other side, in opposite horizontal ends.
        """
        self['A:PP01:E'] = value

    @property
    def m1_photocurrent_signal(self):
        """Return induced voltage in the mirror photocollector [V]."""
        return self['A:RIO01:9215A:ai0']

    @property
    def dvf2_z_pos(self):
        """Return DVF2 base motor longitudinal position [mm]."""
        return self['B:PP01:E.RBV']

    @dvf2_z_pos.setter
    def dvf2_z_pos(self, value):
        """Set DVF2 base motor longitudinal position [mm]."""
        self['B:PP01:E'] = value

    @property
    def dvf2_lens_pos(self):
        """Return DVF2 lens motor position [mm]."""
        return self['B:PP01:F.RBV']

    @dvf2_lens_pos.setter
    def dvf2_lens_pos(self, value):
        """Set DVF2 lens motor position [mm]."""
        self['B:PP01:F'] = value

    @property
    def m1_temperature_ref(self):
        """Return M1 temperature reference setpoint [°C]."""
        return self['A:RIO01:M1_CtrltempSp']

    @m1_temperature_ref.setter
    def m1_temperature_ref(self, value):
        """Set M1 temperature reference [°C]."""
        self['A:RIO01:M1_CtrltempSp'] = value

    @property
    def m1_temperature_0(self):
        """Return M1 temperature at cold finger [°C]."""
        return self['A:RIO01:9226B:temp0']

    @property
    def m1_temperature_1(self):
        """Return M1 temperature at braid mirror [°C]."""
        return self['A:RIO01:9226B:temp1']

    @property
    def m1_temperature_2(self):
        """Return M1 temperature at bar braid cold finger [°C]."""
        return self['A:RIO01:9226B:temp2']

    @property
    def m1_temperature_3(self):
        """Return M1 temperature at peltier cold side [°C]."""
        return self['A:RIO01:9226B:temp3']

    @property
    def m1_temperature_4(self):
        """Return M1 temperature at peltier hot side [°C]."""
        return self['A:RIO01:9226B:temp4']

    @property
    def cmd_m1_roty_stop(self, timeout=None):
        """Stop linear actuator for RotY rotation."""
        return self._cmd_motor_stop('A:PP01:A.STOP', timeout)

    def cmd_m1_tx_stop(self, timeout=None):
        """Stop linear actuator for Tx translation."""
        return self._cmd_motor_stop('A:PP01:B.STOP', timeout)

    def cmd_m1_y1_stop(self, timeout=None):
        """Stop linear actuator Y1."""
        return self._cmd_motor_stop('A:PP01:C.STOP', timeout)

    def cmd_m1_y2_stop(self, timeout=None):
        """Stop linear actuator Y2."""
        return self._cmd_motor_stop('A:PP01:D.STOP', timeout)

    def cmd_m1_y3_stop(self, timeout=None):
        """Stop linear actuator Y3."""
        return self._cmd_motor_stop('A:PP01:E.STOP', timeout)

    def cmd_dvf2_z_stop(self, timeout=None):
        """Stop DVF2 base motor."""
        return self._cmd_motor_stop('B:PP01:E.STOP', timeout)

    def cmd_dvf2_lens_stop(self, timeout=None):
        """Stop DVF2 lens motor."""
        return self._cmd_motor_stop('B:PP01:F.STOP', timeout)

    def _cmd_motor_stop(self, propty, timeout):
        timeout = self._DEFAULT_MOTOR_TIMEOUT if timeout is None else timeout
        self[propty] = 1
        return self.wait(propty, 0, timeout=timeout)



"""Module to deal with correctors."""

import logging as _log
import time as _time
import traceback as _traceback

import numpy as _np

from .. import util as _util
from ..epics import PV as _PV
from ..thread import RepeaterThread as _Repeat
from ..pwrsupply.csdev import Const as _PSConst
from ..timesys.csdev import Const as _TIConst
from ..search import HLTimeSearch as _HLTimesearch
from ..envars import VACA_PREFIX as LL_PREF
from ..namesys import SiriusPVName as _PVName

from .base_class import BaseClass as _BaseClass, \
    BaseTimingConfig as _BaseTimingConfig, compare_kicks as _compare_kicks

TIMEOUT = 0.05
WAIT_CORRS = False


class Corrector(_BaseTimingConfig):
    """Corrector class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name[:2])
        self._name = _PVName(corr_name)
        self._pvs['sp'] = None
        self._pvs['rb'] = None

    @property
    def name(self):
        """Corrector name."""
        return self._name

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        return self.connected

    @property
    def opmode(self):
        """Opmode."""
        return 1

    @opmode.setter
    def opmode(self, val):
        """."""

    @property
    def state(self):
        """State."""
        pv = self._config_pvs_rb["PwrState"]
        return pv.value == _PSConst.PwrStateSel.On if pv.connected else False

    @state.setter
    def state(self, boo):
        """."""
        val = _PSConst.PwrStateSel.On if boo else _PSConst.PwrStateSel.Off
        pv = self._config_pvs_sp["PwrState"]
        self._config_ok_vals["PwrState"] = val
        if self.put_enable and pv.connected:
            pv.put(val, wait=False)

    @property
    def value(self):
        """Value."""
        if self._pvs['rb'].connected:
            return self._pvs['rb'].value

    @value.setter
    def value(self, val):
        """."""
        self._pvs['sp'].put(val, wait=False)

    @property
    def refvalue(self):
        """."""
        return self.value


class RFCtrl(Corrector):
    """RF control class."""

    TINY_VAR = 0.01  # [Hz]
    LARGE_VAR = 10000  # [Hz]
    MAX_DELTA = 200  # [Hz]
    INTVL = 1  # [s]

    def __init__(self, acc):
        """Init method."""
        super().__init__(acc)
        self._name = self._csorb.RF_GEN_NAME
        opt = {"connection_timeout": TIMEOUT}
        pvpref = LL_PREF + ("-" if LL_PREF else "") + self._name
        self._pvs['sp'] = _PV(pvpref + ":GeneralFreq-SP", **opt)
        self._pvs['rb'] = _PV(pvpref + ":GeneralFreq-RB", **opt)
        self._config_ok_vals = {"PwrState": 1}
        self._config_pvs_sp = dict()
        self._config_pvs_rb = dict()

    @property
    def value(self):
        """Value."""
        if self._pvs['rb'].connected:
            return self._pvs['rb'].value

    @value.setter
    def value(self, freq):
        """."""
        freq0 = self.value
        if freq0 is None or freq is None:
            return
        delta = abs(freq - freq0)
        if delta < self.TINY_VAR or delta > self.LARGE_VAR:
            return
        npoints = int(delta // self.MAX_DELTA) + 2
        freq_span = _np.linspace(freq0, freq, npoints)[1:]
        for i, freq in enumerate(freq_span, 1):
            self._pvs['sp'].put(freq, wait=False)
            if i != freq_span.size:
                _time.sleep(self.INTVL)

    @property
    def state(self):
        """State."""
        return True

    @state.setter
    def state(self, boo):
        """."""
        return


class CHCV(Corrector):
    """CHCV class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name)
        opt = {"connection_timeout": TIMEOUT}
        pvsp = self._name.substitute(
            prefix=LL_PREF, propty_name="Kick", propty_suffix="SP"
        )
        pvrb = pvsp.substitute(propty_suffix="RB")
        pvref = pvsp.substitute(propty_name="KickRef", propty_suffix="Mon")
        self._pvs['sp'] = _PV(pvsp, **opt)
        self._pvs['rb'] = _PV(pvrb, **opt)
        self._pvs['ref'] = _PV(pvref, **opt)

        pvoffwfmsp = self._name.substitute(
            prefix=LL_PREF, propty_name='WfmOffsetKick', propty_suffix='SP')
        pvoffwfmrb = self._name.substitute(
            prefix=LL_PREF, propty_name='WfmOffsetKick', propty_suffix='RB')
        self._pvs['wfm_offset_sp'] = _PV(pvoffwfmsp)
        self._pvs['wfm_offset_rb'] = _PV(pvoffwfmrb)
        self._config_ok_vals = {
            'OpMode': _PSConst.OpMode.SlowRef,
            'PwrState': _PSConst.PwrStateSel.On}
        pvpref = self._name.substitute(prefix=LL_PREF)
        self._config_pvs_sp = {
            'OpMode': _PV(pvpref.substitute(propty='OpMode-Sel'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sel'), **opt)}
        self._config_pvs_rb = {
            'OpMode': _PV(pvpref.substitute(propty='OpMode-Sts'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sts'), **opt)}

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        isok = self.opmode == self._config_ok_vals["OpMode"]
        if not isok:
            msg = "OpMode not Ok {0:s}".format(self.name)
            _log.warning(msg)
        return isok

    @property
    def opmode(self):
        """Opmode."""
        pvobj = self._config_pvs_rb["OpMode"]
        if not pvobj.connected:
            return None
        elif pvobj.value == _PSConst.States.SlowRefSync:
            return _PSConst.OpMode.SlowRefSync
        elif pvobj.value == _PSConst.States.SlowRef:
            return _PSConst.OpMode.SlowRef
        elif pvobj.value == _PSConst.States.RmpWfm:
            return _PSConst.OpMode.RmpWfm
        return pvobj.value

    @opmode.setter
    def opmode(self, val):
        """."""
        pvobj = self._config_pvs_sp["OpMode"]
        self._config_ok_vals["OpMode"] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def value(self):
        """Value."""
        if not self._pvs['rb'].connected:
            return None
        if self._config_ok_vals['OpMode'] == _PSConst.OpMode.RmpWfm:
            return self.wfm_offset_kick
        return self._pvs['rb'].value

    @value.setter
    def value(self, val):
        """."""
        if self._config_ok_vals['OpMode'] == _PSConst.OpMode.RmpWfm:
            self.wfm_offset_kick = val
        else:
            self._pvs['sp'].put(val, wait=False)

    @property
    def wfm_offset_kick(self):
        """."""
        if not self._pvs['wfm_offset_rb'].connected:
            return None
        return self._pvs['wfm_offset_rb'].value

    @wfm_offset_kick.setter
    def wfm_offset_kick(self, val):
        """."""
        self._pvs['wfm_offset_sp'].put(val, wait=False)

    @property
    def refvalue(self):
        """."""
        if not self._pvs['ref'].connected:
            return None
        if self._config_ok_vals['OpMode'] == _PSConst.OpMode.RmpWfm:
            return self.wfm_offset_kick
        return self._pvs['ref'].value

    def configure(self):
        """Configure method."""
        if not self.connected:
            return False

        if not self.put_enable:
            return True
        for k, pvo in self._config_pvs_sp.items():
            if k in self._config_ok_vals:
                pvo.put(self._config_ok_vals[k], wait=False)
        return True


class Septum(Corrector):
    """Septum class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name)
        opt = {"connection_timeout": TIMEOUT}
        pvsp = self._name.substitute(
            prefix=LL_PREF, propty_name="Kick", propty_suffix="SP"
        )
        pvrb = pvsp.substitute(propty_suffix="RB")
        self._pvs['sp'] = _PV(pvsp, **opt)
        self._pvs['rb'] = _PV(pvrb, **opt)
        self._nominalkick = 98.55  # mrad
        self._config_ok_vals = {
            "Pulse": 1,
            "PwrState": _PSConst.PwrStateSel.On,
        }
        pvpref = self._name.substitute(prefix=LL_PREF)
        self._config_pvs_sp = {
            "Pulse": _PV(pvpref.substitute(propty="Pulse-Sel"), **opt),
            "PwrState": _PV(pvpref.substitute(propty="PwrState-Sel"), **opt),
        }
        self._config_pvs_rb = {
            "Pulse": _PV(pvpref.substitute(propty="Pulse-Sts"), **opt),
            "PwrState": _PV(pvpref.substitute(propty="PwrState-Sts"), **opt),
        }

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        return self.opmode == self._config_ok_vals["Pulse"]

    @property
    def opmode(self):
        """Opmode."""
        pv = self._config_pvs_rb["Pulse"]
        if not pv.connected:
            return None
        return pv.value

    @opmode.setter
    def opmode(self, val):
        """."""
        pv = self._config_pvs_sp["Pulse"]
        self._config_ok_vals["Pulse"] = val
        if self.put_enable and pv.connected and pv.value != val:
            pv.put(val, wait=False)

    @property
    def value(self):
        """Value."""
        if self._pvs['rb'].connected:
            val = self._pvs['rb'].value
            if val is not None:
                return -(val + self._nominalkick) * 1e3

    @value.setter
    def value(self, val):
        """."""
        val = val / 1e3 + self._nominalkick
        self._pvs['sp'].put(-val, wait=False)


def get_corr(name):
    """."""
    if name.dis == "PU":
        return Septum(name)
    else:
        return CHCV(name)


class TimingConfig(_BaseTimingConfig):
    """Timing configuration class."""

    def __init__(self, acc):
        """Init method."""
        super().__init__(acc)
        evt = self._csorb.evt_cor_name
        pref = LL_PREF + ("-" if LL_PREF else "")
        pref_name = pref + self._csorb.evg_name + ":" + evt
        trig = self._csorb.trigger_cor_name
        opt = {"connection_timeout": TIMEOUT}
        self._pvs['evt_sender'] = _PV(pref_name + "ExtTrig-Cmd", **opt)
        src_val = self._csorb.CorrExtEvtSrc._fields.index(evt)
        src_val = self._csorb.CorrExtEvtSrc[src_val]
        self.EVT = src_val

        src_val = self._csorb.CorrExtEvtSrc._fields.index(
            self._csorb.evt_rmpbo
        )
        src_val = self._csorb.CorrExtEvtSrc[src_val]
        self.RMPBO = src_val

        src_val = self._csorb.CorrExtEvtSrc._fields.index(
            self._csorb.clk_cor_name
        )
        src_val = self._csorb.CorrExtEvtSrc[src_val]
        self.CLK = src_val

        self._config_ok_vals = {
            "Mode": _TIConst.EvtModes.External,
            "Src": src_val,
            "NrPulses": 1,
            "Duration": 150.0,
            "State": 1,
            "Polarity": 0,
        }
        if _HLTimesearch.has_delay_type(trig):
            self._config_ok_vals["RFDelayType"] = _TIConst.TrigDlyTyp.Manual
        pref_trig = pref + trig + ":"
        self._config_pvs_rb = {
            "Mode": _PV(pref_name + "Mode-Sts", **opt),
            "Src": _PV(pref_trig + "Src-Sts", **opt),
            "DelayRaw": _PV(pref_trig + "DelayRaw-RB", **opt),
            "DeltaDelayRaw": _PV(pref_trig + "DeltaDelayRaw-RB", **opt),
            "NrPulses": _PV(pref_trig + "NrPulses-RB", **opt),
            "Duration": _PV(pref_trig + "Duration-RB", **opt),
            "State": _PV(pref_trig + "State-Sts", **opt),
            "Polarity": _PV(pref_trig + "Polarity-Sts", **opt),
        }
        self._config_pvs_sp = {
            "Mode": _PV(pref_name + "Mode-Sel", **opt),
            "Src": _PV(pref_trig + "Src-Sel", **opt),
            "DelayRaw": _PV(pref_trig + "DelayRaw-SP", **opt),
            "DeltaDelayRaw": _PV(pref_trig + "DeltaDelayRaw-SP", **opt),
            "NrPulses": _PV(pref_trig + "NrPulses-SP", **opt),
            "Duration": _PV(pref_trig + "Duration-SP", **opt),
            "State": _PV(pref_trig + "State-Sel", **opt),
            "Polarity": _PV(pref_trig + "Polarity-Sel", **opt),
        }
        if _HLTimesearch.has_delay_type(trig):
            self._config_pvs_rb["RFDelayType"] = _PV(
                pref_trig + "RFDelayType-Sts", **opt
            )
            self._config_pvs_sp["RFDelayType"] = _PV(
                pref_trig + "RFDelayType-Sel", **opt
            )

    def send_evt(self):
        """Send event method."""
        self._pvs['evt_sender'].value = 1

    @property
    def state(self):
        """State."""
        pv = self._config_pvs_rb['State']
        return pv.value if pv.connected else 0

    @state.setter
    def state(self, val):
        """."""
        pv = self._config_pvs_sp['State']
        self._config_ok_vals['State'] = val
        if self.put_enable and pv.connected:
            pv.put(val, wait=False)

    @property
    def sync_type(self):
        """."""
        return self._config_pvs_rb["Src"].value

    @sync_type.setter
    def sync_type(self, value):
        """."""
        if value not in (self.EVT, self.CLK, self.RMPBO):
            return
        self._config_ok_vals["Src"] = value
        pvobj = self._config_pvs_sp["Src"]
        if self.put_enable and pvobj.connected:
            pvobj.value = value

    @property
    def delayraw(self):
        """."""
        defv = 0
        pvobj = self._config_pvs_rb["DelayRaw"]
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @delayraw.setter
    def delayraw(self, value):
        """."""
        # Make sure Delta Delay is zero:
        pvobj = self._config_pvs_sp["DeltaDelayRaw"]
        if self.put_enable and pvobj.connected:
            pvobj.value *= 0

        pvobj = self._config_pvs_sp["DelayRaw"]
        if self.put_enable and pvobj.connected:
            pvobj.value = int(value)


class BaseCorrectors(_BaseClass):
    """Base correctors class."""


class EpicsCorrectors(BaseCorrectors):
    """Class to deal with correctors."""

    TINY_INTERVAL = 0.005
    NUM_TIMEOUT = 1000
    MAX_PROB = 5
    ACQRATE = 2

    def __init__(self, acc, prefix='', callback=None):
        """Initialize the instance."""
        super().__init__(acc, prefix=prefix, callback=callback)
        self._sofb = None

        self._names = self._csorb.ch_names + self._csorb.cv_names
        self._corrs = [get_corr(dev) for dev in self._names]
        if self.isring:
            self._corrs.append(RFCtrl(self.acc))

        if self.acc == 'SI':
            self.sync_kicks = self._csorb.CorrSync.Off
            self.timing = TimingConfig(acc)
        self._corrs_thread = _Repeat(
            1 / self.ACQRATE, self._update_corrs_strength, niter=0
        )
        self._corrs_thread.start()

    @property
    def connected(self):
        """."""
        for cor in self._corrs:
            if not cor.connected:
                return False
        if self.acc == 'SI' and not self.timing.connected:
            return False
        return True

    @property
    def sofb(self):
        """."""
        return self._sofb

    @sofb.setter
    def sofb(self, sofb):
        self._sofb = sofb

    @property
    def corrs(self):
        """."""
        return self._corrs

    def wait_for_connection(self, timeout=10):
        """."""
        t0_ = _time.time()
        for cor in self._corrs:
            tout = timeout - (_time.time() - t0_)
            if tout <= 0 or not cor.wait_for_connection(tout):
                return False
        if self.acc != 'SI':
            return True
        tout = timeout - (_time.time() - t0_)
        if tout <= 0 or not self.timing.wait_for_connection(tout):
            return False
        return True

    def shutdown(self):
        """Shutdown threads."""
        self._corrs_thread.resume()
        self._corrs_thread.stop()
        self._corrs_thread.join()

    def get_map2write(self):
        """Get the write methods of the class."""
        dbase = {"CorrConfig-Cmd": self.configure_correctors}
        if self.acc == "SI":
            dbase["CorrSync-Sel"] = self.set_corrs_mode
        return dbase

    def apply_kicks(self, values):
        """Apply kicks.

        Will return -2 if there is a problem with some PS.
        Will return -1 if PS did not finish applying last kick.
        Will return 0 if all previous kick were implemented.
        Will return >0 indicating how many previous kicks were not implemented.
        """
        if self.acc == "BO":
            msg = "ERR: Cannot correct Orbit in Booster. Use Ramp Interface!"
            self._update_log(msg)
            _log.error(msg[5:])
            return 0

        strn = '    TIMEIT: {0:20s} - {1:7.3f}'
        _log.debug('    TIMEIT: BEGIN')
        time1 = _time.time()

        not_nan_idcs = ~_np.isnan(values)
        # Send correctors setpoint
        for i, corr in enumerate(self._corrs):
            if not_nan_idcs[i]:
                self.put_value_in_corr(corr, values[i])
        time2 = _time.time()
        _log.debug(strn.format('send sp:', 1000*(time2-time1)))

        # Wait for readbacks to be updated
        if WAIT_CORRS and self._timed_out(values, mode='ready'):
            return -1
        time3 = _time.time()
        _log.debug(strn.format('check ready:', 1000*(time3-time2)))

        # Send trigger signal for implementation
        self.send_evt()
        time4 = _time.time()
        _log.debug(strn.format('send evt:', 1000*(time4-time3)))

        # Wait for references to be updated
        if WAIT_CORRS:
            self._timed_out(values, mode='applied')
        time5 = _time.time()
        _log.debug(strn.format('check applied:', 1000*(time5-time4)))
        _log.debug('    TIMEIT: END')
        return 0

    def put_value_in_corr(self, corr, value):
        """Put value in corrector method."""
        if not corr.connected:
            msg = "ERR: " + corr.name + " not connected."
            self._update_log(msg)
            _log.error(msg[5:])
        elif not corr.state:
            msg = "ERR: " + corr.name + " is off."
            self._update_log(msg)
            _log.error(msg[5:])
        elif not corr.opmode_ok:
            msg = "ERR: " + corr.name + " mode not configured."
            self._update_log(msg)
            _log.error(msg[5:])
        else:
            corr.value = value

    def send_evt(self):
        """Send event method."""
        if self.acc != "SI" or self.sync_kicks != self._csorb.CorrSync.Event:
            return
        if not self.timing.connected:
            msg = "ERR: timing disconnected."
            self._update_log(msg)
            _log.error(msg[5:])
            return
        elif not self.timing.is_ok:
            msg = "ERR: timing not configured."
            self._update_log(msg)
            _log.error(msg[5:])
            return
        self.timing.send_evt()

    def get_strength(self):
        """Get the correctors strengths."""
        corr_values = _np.zeros(self._csorb.nr_corrs, dtype=float)

        for i, corr in enumerate(self._corrs):
            if corr.connected and corr.value is not None:
                corr_values[i] = corr.refvalue
            else:
                msg = "ERR: Failed to get value from "
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
        return corr_values

    def _update_corrs_strength(self):
        try:
            corr_vals = self.get_strength()
            self.run_callbacks("KickCH-Mon", corr_vals[: self._csorb.nr_ch])
            self.run_callbacks(
                "KickCV-Mon",
                corr_vals[self._csorb.nr_ch : self._csorb.nr_chcv],
            )
            if self.isring and corr_vals[-1] > 0:
                # NOTE: I have to check whether the RF frequency is larger
                # than zero not to take the inverse of 0. It will be 0 in case
                # there is a failure to get the RF frequency from its PV.
                rfv = corr_vals[-1]
                circ = 1 / rfv * self._csorb.harm_number * 299792458
                self.run_callbacks("KickRF-Mon", rfv)
                self.run_callbacks("OrbLength-Mon", circ)
        except Exception as err:
            self._update_log("ERR: " + str(err))
            _log.error(_traceback.format_exc())

    def set_corrs_mode(self, value):
        """Set mode of CHs and CVs method. Only called when acc==SI."""
        if value not in self._csorb.CorrSync:
            return False

        self.timing.state = 0
        self.sync_kicks = value
        refs = [c.value for c in self._corrs[:-1]]
        val = _PSConst.OpMode.SlowRefSync
        if value == self._csorb.CorrSync.Off:
            self.set_timing_delay(0)
            val = _PSConst.OpMode.SlowRef
        elif value == self._csorb.CorrSync.Event:
            self.set_timing_delay(0)
            self.timing.sync_type = self.timing.EVT
        elif value == self._csorb.CorrSync.Clock:
            self.set_timing_delay(self._csorb.CORR_DEF_DELAY)
            self.timing.sync_type = self.timing.CLK
        elif value == self._csorb.CorrSync.RmpBO:
            self.set_timing_delay(0)
            self.timing.sync_type = self.timing.RMPBO
            val = _PSConst.OpMode.RmpWfm

        # Time to wait any waveform to finish
        _time.sleep(0.6)

        mask = self._get_mask()
        for i, corr in enumerate(self._corrs[:-1]):
            if mask[i] and not corr.connected:
                msg = "ERR: Failed to configure "
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            corr.put_enable = mask[i]
            corr.opmode = val
            corr.value = refs[i]

        self.timing.state = 1
        strsync = self._csorb.CorrSync._fields[self.sync_kicks]
        msg = "Synchronization set to {0:s}".format(strsync)
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks("CorrSync-Sts", value)
        return True

    def set_timing_delay(self, value):
        """."""
        frf = 499666 / 4  # [kHz]
        raw = int(value * frf)
        self.timing.delayraw = raw
        return True

    def configure_correctors(self, val):
        """Configure correctors method."""
        _ = val
        corrs = self._get_used_corrs(include_rf=True)
        for corr in corrs:
            if not corr.connected:
                msg = "ERR: Failed to configure "
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
                continue
            # Do not configure opmode in BO corrs because they are ramping.
            if self.acc == "BO":
                corr.state = True
                continue
            corr.configure()
        if not self.isring:
            return True
        if self.acc == "SI" and self.sync_kicks != self._csorb.CorrSync.Off:
            if not self.timing.configure():
                msg = "ERR: Failed to configure timing"
                self._update_log(msg)
                _log.error(msg[5:])
        return True

    def _update_status(self):
        status = 0b0000111
        if self.acc == "SI":
            status = 0b1111111
        elif self.isring:
            status = 0b0011111

        chcvs = self._get_used_corrs(include_rf=False)

        status = _util.update_bit(
            status,
            bit_pos=0,
            bit_val=not all(corr.connected for corr in chcvs),
        )
        # Do not check mode of BO correctors because they are ramping.
        opmode_ok = True
        if self.acc != "BO":
            opmode_ok = all(corr.opmode_ok for corr in chcvs)
        status = _util.update_bit(status, bit_pos=1, bit_val=not opmode_ok)
        status = _util.update_bit(
            status, bit_pos=2, bit_val=not all(corr.state for corr in chcvs)
        )
        if self.acc == "SI" and self.sync_kicks != self._csorb.CorrSync.Off:
            tim_conn = self.timing.connected
            tim_conf = self.timing.is_ok
        else:
            tim_conn = tim_conf = True
        status = _util.update_bit(status, bit_pos=3, bit_val=not tim_conn)
        status = _util.update_bit(status, bit_pos=4, bit_val=not tim_conf)
        if self.isring:
            rfctrl = self._corrs[-1]
            status = _util.update_bit(
                status, bit_pos=5, bit_val=not rfctrl.connected
            )
            status = _util.update_bit(
                status, bit_pos=6, bit_val=not rfctrl.state
            )
        self._status = status
        self.run_callbacks("CorrStatus-Mon", status)

    def _timed_out(self, values, mode="ready"):
        okg = [False] * len(self._corrs)
        for _ in range(self.NUM_TIMEOUT):
            for i, corr in enumerate(self._corrs):
                if okg[i]:
                    continue
                if _np.isnan(values[i]):
                    okg[i] = True
                    continue
                val = corr.value if mode == "ready" else corr.refvalue
                if val is None:
                    continue
                if isinstance(corr, RFCtrl):
                    okg[i] = _compare_kicks(
                        values[i], val, atol=RFCtrl.TINY_VAR
                    )
                else:
                    okg[i] = _compare_kicks(values[i], val)
            if all(okg):
                return False
            _time.sleep(self.TINY_INTERVAL)

        self._print_guilty(okg, mode=mode)
        return True

    def _print_guilty(
        self, okg, mode="ready", fret=None, currs=None, refs=None
    ):
        msg_tmpl = "ERR: timeout {0:3s}: {1:s}"
        data = [tuple()] * len(self._corrs)
        if mode == "prob_code":
            msg_tmpl = "ERR: {0:s} --> {1:s}: code={2:d}"
            data = zip(fret)
        elif mode == "prob_curr":
            msg_tmpl = "ERR: {0:s} --> {1:s}: curr={2:.4g}, ref={3:.4g}"
            data = zip(currs, refs)
        elif mode == "diff":
            msg_tmpl = "ERR: Corrector {1:s} diff from setpoint!"
        for oki, corr, args in zip(okg, self._corrs, data):
            if not oki:
                msg = msg_tmpl.format(mode, corr.name, *args)
                self._update_log(msg)
                _log.error(msg[5:])

    def _get_mask(self):
        if self.sofb is not None and self.sofb.matrix is not None:
            mask = self.sofb.matrix.corrs_enbllist
        else:
            mask = _np.ones(len(self._corrs), dtype=bool)
        return mask

    def _get_used_corrs(self, include_rf=False):
        corrs = []
        nrc = None if include_rf else self._csorb.nr_chcv
        mask = self._get_mask()[:nrc]
        for i, corr in enumerate(self._corrs[:nrc]):
            corr.put_enable = mask[i]
            if mask[i]:
                corrs.append(corr)
        return corrs



"""."""

from .blm_search import BLMSearch
from .bpms_search import BPMSearch
from .gamma_monitor_search import GammaMonitorSearch
from .hl_time_search import HLTimeSearch
from .id_search import IDSearch
from .ioc_search import IOCSearch
from .ll_time_search import LLTimeSearch
from .ma_search import MASearch
from .orb_intlk_search import OrbIntlkSearch
from .ps_search import PSSearch
from .rabpm_search import RaBPMSearch

del blm_search
del bpms_search
del gamma_monitor_search
del hl_time_search
del id_search
del ioc_search
del ll_time_search
del ma_search
del orb_intlk_search
del ps_search
del rabpm_search



"""Define PVs, constants and properties of Filling Pattern SoftIOCs."""

from .. import csdev as _csdev

# --- Const class ---


class Const(_csdev.Const):
    """Const class defining Filling Pattern constants."""

    FP_MAX_ARR_SIZE = 40000
    FP_HARM_NUM = 864
    FP_MAX_UPDT_TIME = 600  # [s]
    DEVNAME = 'SI-Glob:DI-FPMOsc'


_c = Const  # syntactic sugar


# --- Databases ---

def get_si_fpmosc_database():
    """Return SI FPMOsc Soft IOC database."""
    pvs_db = {'Version-Cte': {'type': 'string', 'value': 'UNDEF'}}

    pvs_db['UpdateTime-SP'] = {
        'type': 'int', 'value': 5, 'unit': 's',
        'low': 0, 'high': _c.FP_MAX_UPDT_TIME,
        'lolo': 0, 'hihi': _c.FP_MAX_UPDT_TIME,
        'lolim': 0, 'hilim': _c.FP_MAX_UPDT_TIME,
    }
    pvs_db['UpdateTime-RB'] = {
        'type': 'int', 'value': 5, 'unit': 's',
        'low': 0, 'high': _c.FP_MAX_UPDT_TIME,
        'lolo': 0, 'hihi': _c.FP_MAX_UPDT_TIME,
        'lolim': 0, 'hilim': _c.FP_MAX_UPDT_TIME,
    }
    pvs_db['FiducialOffset-SP'] = {
        'type': 'int', 'value': 0, 'unit': 'bunches',
        'low': -_c.FP_HARM_NUM, 'high': _c.FP_HARM_NUM,
        'lolo': -_c.FP_HARM_NUM, 'hihi': _c.FP_HARM_NUM,
        'lolim': -_c.FP_HARM_NUM, 'hilim': _c.FP_HARM_NUM,
    }
    pvs_db['FiducialOffset-RB'] = {
        'type': 'int', 'value': 0, 'unit': 'bunches',
        'low': -_c.FP_HARM_NUM, 'high': _c.FP_HARM_NUM,
        'lolo': -_c.FP_HARM_NUM, 'hihi': _c.FP_HARM_NUM,
        'lolim': -_c.FP_HARM_NUM, 'hilim': _c.FP_HARM_NUM,
    }
    pvs_db['TimeOffset-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 4, 'unit': 'ns'
    }
    pvs_db['FillPatternRef-SP'] = {
        'type': 'float', 'prec': 5, 'count': _c.FP_HARM_NUM,
        'value': [0.0, ] * _c.FP_HARM_NUM, 'unit': 'rel'
    }
    pvs_db['FillPatternRef-RB'] = {
        'type': 'float', 'prec': 5, 'count': _c.FP_HARM_NUM,
        'value': [0.0, ] * _c.FP_HARM_NUM, 'unit': 'rel'
    }

    pvs_db['UniFillEqCurrent-Mon'] = {
        'type': 'float', 'value': 0, 'unit': 'mA', 'prec': 2,
    }
    pvs_db['ErrorRelStd-Mon'] = {
        'type': 'float', 'value': 0, 'unit': '%', 'prec': 2,
    }
    pvs_db['ErrorKLDiv-Mon'] = {
        'type': 'float', 'value': 0, 'unit': '', 'prec': 5,
    }
    pvs_db['FillPatternRef-Mon'] = {
        'type': 'float', 'prec': 5, 'count': _c.FP_HARM_NUM,
        'value': [0.0, ] * _c.FP_HARM_NUM, 'unit': 'mA'
    }
    pvs_db['FillPattern-Mon'] = {
        'type': 'float', 'prec': 5, 'count': _c.FP_HARM_NUM,
        'value': [0.0, ] * _c.FP_HARM_NUM, 'unit': 'mA'
    }
    pvs_db['Time-Mon'] = {
        'type': 'float', 'prec': 3, 'count': _c.FP_HARM_NUM,
        'value': [0.0, ] * _c.FP_HARM_NUM, 'unit': 'ns'
    }
    pvs_db['Raw-Mon'] = {
        'type': 'float', 'prec': 3, 'count': _c.FP_MAX_ARR_SIZE,
        'value': [0.0, ] * _c.FP_MAX_ARR_SIZE, 'unit': 'mA'
    }
    pvs_db['RawAmp-Mon'] = {
        'type': 'float', 'prec': 3, 'count': _c.FP_MAX_ARR_SIZE,
        'value': [0.0, ] * _c.FP_MAX_ARR_SIZE, 'unit': 'mA'
    }
    pvs_db['RawTime-Mon'] = {
        'type': 'float', 'prec': 3, 'count': _c.FP_MAX_ARR_SIZE,
        'value': [0.0, ] * _c.FP_MAX_ARR_SIZE, 'unit': 'ns'
    }
    return _csdev.add_pvslist_cte(pvs_db)



"""Load and process RaBPM data from static table in remote server."""

from threading import Lock as _Lock
from siriuspy.namesys import Filter as _Filter, SiriusPVName as _PVName
import siriuspy.clientweb as _web

_timeout = 1.0


class RaBPMSearch:
    """Class with mapping devices in uTCA crates.

    Data are read from the Control System Constants Server.
    """

    _mapping = None

    _lock = _Lock()

    @classmethod
    def reset(cls):
        """Reload data from files."""
        cls._mapping = None
        cls._get_data()

    @classmethod
    def server_online(cls):
        """Return True/False if Sirius web server is online."""
        return _web.server_online()

    @classmethod
    def get_names(cls, filters=None):
        """Return a list with the device names for the given filter.

        Parameters
        ----------
        filters : Filter object, optional
            Filter select a subset of devices.
        """
        cls._get_data()
        return _Filter.process_filters(cls._names, filters=filters)

    @classmethod
    def get_slots(cls, names=None, filters=None):
        """Return a list with the slots of the given devices.

        Parameters
        ----------
        names : string list, optional
            Desired device name.
        filters : Filter object, optional
            Filter select a subset of devices.
        """
        cls._get_data()
        if not names:
            names = cls.get_names(filters=filters)
        return [cls._mapping[k]["slot"] for k in names]

    @classmethod
    def get_crate_names(cls, names=None, filters=None):
        """Return a list with the crate names of the given devices.

        Parameters
        ----------
        names : string list, optional
            Desired device name.
        filters : Filter object, optional
            Filter select a subset of devices.
        """
        cls._get_data()
        if not names:
            names = cls.get_names(filters=filters)
        return [cls._mapping[k]["crate_name"] for k in names]

    @classmethod
    def get_crate_numbers(cls, names=None, filters=None):
        """Return a list with the crate numbers of the given devices.

        Parameters
        ----------
        names : string list, optional
            Desired device name.
        filters : Filter object, optional
            Filter select a subset of devices.
        """
        cls._get_data()
        if not names:
            names = cls.get_names(filters=filters)
        return [cls._mapping[k]["crate_number"] for k in names]

    @classmethod
    def conv_devname_2_slot(cls, name):
        """Return the slot of the given device name.

        Parameters
        ----------
        names : string
            Desired device name.
        """
        cls._get_data()
        return cls._mapping[name]["slot"]

    @classmethod
    def conv_devname_2_cratename(cls, name):
        """Return the crate name of the given device name.

        Parameters
        ----------
        names : string
            Desired device name.
        """
        cls._get_data()
        return cls._mapping[name]["crate_name"]

    @classmethod
    def conv_devname_2_cratenumber(cls, name):
        """Return the crate number of the given device name.

        Parameters
        ----------
        names : string
            Desired device name.
        """
        cls._get_data()
        return cls._mapping[name]["crate_number"]

    @classmethod
    def get_bpm_pair(cls, name):
        """ Each BPM share its slot with another BPM, defining a BPM pair.
        Return the pair of a given BPM name.

        Parameters
        ----------
        names : string
            Desired device name.
        """

        slot = cls.conv_devname_2_slot(name)
        cratename = cls.conv_devname_2_cratename(name)

        target = {"slot": slot, "crate_name": cratename}

        doubles = [key for key, val in cls._mapping.items() if
                   (val.get("slot") == target.get("slot") and val.get("crate_name") == target.get("crate_name"))]

        doubles.remove(name)
        if (len(doubles) != 1):
            print("Device does not have a pair.")
            return

        return doubles[0]

    @classmethod
    def _get_data(cls):
        with cls._lock:
            if cls._mapping is not None:
                return
            if not _web.server_online():
                raise Exception('could not read data from web server!')
            cls._build_mapping()

    @classmethod
    def _build_mapping(cls):
        data = _web.crates_mapping(timeout=_timeout)
        mapping = dict()
        for line in data.splitlines():
            line = line.strip()
            if not line or line[0] == '#':
                continue  # empty line
            dev, slot, *_, crate = line.split()
            dev = _PVName(dev)
            if dev in mapping.keys():
                raise Exception('BPM {0:s} double entry.'.format(dev))
            crate_num = crate[-2:]
            crate_name = ('IA-20RaBPMTL' if int(crate_num) == 21 else
                          f'IA-{crate_num}RaBPM')
            mapping[dev] = {
                "slot": int(slot),
                "crate_name": crate_name,
                "crate_number": int(crate_num)}
        cls._mapping = mapping
        cls._names = sorted(cls._mapping.keys())
